import gzip
import mimetypes
import os
import re
from datetime import datetime
from functools import wraps
from io import BytesIO, UnsupportedOperation
from time import time

import magic
from django.core.exceptions import ImproperlyConfigured
from django.core.files import File
from django.core.files.storage import Storage
from six.moves.urllib import parse as urlparse

from swift.utils import setting

try:
    from django.utils.deconstruct import deconstructible
except ImportError:
    def deconstructible(arg):
        return arg

try:
    import swiftclient
    from swiftclient.utils import generate_temp_url
except ImportError:
    raise ImproperlyConfigured("Could not load swiftclient library")


def validate_settings(backend):
    # Check mandatory parameters
    if not backend.api_auth_url:
        raise ImproperlyConfigured("The SWIFT_AUTH_URL setting is required")

    if not backend.api_username:
        raise ImproperlyConfigured("The SWIFT_USERNAME setting is required")

    if not backend.api_key:
        raise ImproperlyConfigured("The SWIFT_KEY or SWIFT_PASSWORD setting is required")

    if not backend.container_name:
        raise ImproperlyConfigured("No container name defined. Use SWIFT_CONTAINER_NAME \
        or SWIFT_STATIC_CONTAINER_NAME depending on the backend")

    # Detect auth version if not defined
    # http://docs.openstack.org/developer/python-swiftclient/cli.html#authentication
    if not backend.auth_version:
        if (backend.user_domain_name or backend.user_domain_id) and \
           (backend.project_domain_name or backend.project_domain_id):
            # Set version 3 if domain and project scoping is defined
            backend.auth_version = '3'
        else:
            if backend.tenant_name or backend.tenant_id:
                # Set version 2 if a tenant is defined
                backend.auth_version = '2'
            else:
                # Set version 1 if no tenant is not defined
                backend.auth_version = '1'

    # Enforce auth_version into a string (more future proof)
    backend.auth_version = str(backend.auth_version)

    # Validate v2 auth parameters
    if backend.auth_version == '2':
        if not (backend.tenant_name or backend.tenant_id):
            raise ImproperlyConfigured("SWIFT_TENANT_ID or SWIFT_TENANT_NAME must \
             be defined when using version 2 auth")

    # Validate v3 auth parameters
    if backend.auth_version == '3':
        if not (backend.user_domain_name or backend.user_domain_id):
            raise ImproperlyConfigured("SWIFT_USER_DOMAIN_NAME or \
            SWIFT_USER_DOMAIN_ID must be defined when using version 3 auth")

        if not (backend.project_domain_name or backend.project_domain_id):
            raise ImproperlyConfigured("SWIFT_PROJECT_DOMAIN_NAME or \
            SWIFT_PROJECT_DOMAIN_ID must be defined when using version 3 auth")

        if not (backend.tenant_name or backend.tenant_id):
            raise ImproperlyConfigured("SWIFT_PROJECT_ID or SWIFT_PROJECT_NAME must \
             be defined when using version 3 auth")

    # Validate temp_url parameters
    if backend.use_temp_urls:
        if backend.temp_url_key is None:
            raise ImproperlyConfigured("SWIFT_TEMP_URL_KEY must be set when \
             SWIFT_USE_TEMP_URL is True")

        # Encode temp_url_key as bytes
        try:
            backend.temp_url_key = backend.temp_url_key.encode('ascii')
        except UnicodeEncodeError:
            raise ImproperlyConfigured("SWIFT_TEMP_URL_KEY must ascii")

    # Misc sanity checks
    if not isinstance(backend.os_extra_options, dict):
        raise ImproperlyConfigured("SWIFT_EXTRA_OPTIONS must be a dict")


def prepend_name_prefix(func):
    """
    Decorator that wraps instance methods to prepend the instance's filename
    prefix to the beginning of the referenced filename. Must only be used on
    instance methods where the first parameter after `self` is `name` or a
    comparable parameter of a different name.
    """
    @wraps(func)
    def prepend_prefix(self, name, *args, **kwargs):
        name = self.name_prefix + name
        return func(self, name, *args, **kwargs)
    return prepend_prefix


@deconstructible
class SwiftStorage(Storage):
    api_auth_url = setting('SWIFT_AUTH_URL')
    api_username = setting('SWIFT_USERNAME')
    api_key = setting('SWIFT_KEY') or setting('SWIFT_PASSWORD')
    auth_version = setting('SWIFT_AUTH_VERSION')
    tenant_name = setting('SWIFT_TENANT_NAME') or setting('SWIFT_PROJECT_NAME')
    tenant_id = setting('SWIFT_TENANT_ID') or setting('SWIFT_PROJECT_ID')
    user_domain_name = setting('SWIFT_USER_DOMAIN_NAME')
    user_domain_id = setting('SWIFT_USER_DOMAIN_ID')
    project_domain_name = setting('SWIFT_PROJECT_DOMAIN_NAME')
    project_domain_id = setting('SWIFT_PROJECT_DOMAIN_ID')
    region_name = setting('SWIFT_REGION_NAME')
    container_name = setting('SWIFT_CONTAINER_NAME')
    auto_create_container = setting('SWIFT_AUTO_CREATE_CONTAINER', False)
    auto_create_container_public = setting(
        'SWIFT_AUTO_CREATE_CONTAINER_PUBLIC', False)
    auto_create_container_allow_orgin = setting(
        'SWIFT_AUTO_CREATE_CONTAINER_ALLOW_ORIGIN')
    auto_base_url = setting('SWIFT_AUTO_BASE_URL', True)
    override_base_url = setting('SWIFT_BASE_URL')
    use_temp_urls = setting('SWIFT_USE_TEMP_URLS', False)
    temp_url_key = setting('SWIFT_TEMP_URL_KEY')
    temp_url_duration = setting('SWIFT_TEMP_URL_DURATION', 30 * 60)
    auth_token_duration = setting('SWIFT_AUTH_TOKEN_DURATION', 60 * 60 * 23)
    os_extra_options = setting('SWIFT_EXTRA_OPTIONS', {})
    auto_overwrite = setting('SWIFT_AUTO_OVERWRITE', False)
    lazy_connect = setting('SWIFT_LAZY_CONNECT', False)
    content_type_from_fd = setting('SWIFT_CONTENT_TYPE_FROM_FD', False)
    content_length_from_fd = setting('SWIFT_CONTENT_LENGTH_FROM_FD', True)
    gzip_content_types = setting('SWIFT_GZIP_CONTENT_TYPES', [])
    gzip_unknown_content_type = setting('SWIFT_GZIP_UNKNOWN_CONTENT_TYPE', False)
    gzip_compression_level = setting('SWIFT_GZIP_COMPRESSION_LEVEL', 4)
    _token_creation_time = 0
    _token = ''
    _swift_conn = None
    _base_url = None
    name_prefix = setting('SWIFT_NAME_PREFIX', '')
    full_listing = setting('SWIFT_FULL_LISTING', True)
    max_retries = setting('SWIFT_MAX_RETRIES', 5)
    cache_headers = setting('SWIFT_CACHE_HEADERS', False)

    def __init__(self, **settings):
        # check if some of the settings provided as class attributes
        # should be overwritten
        for name, value in settings.items():
            if hasattr(self, name):
                setattr(self, name, value)

        validate_settings(self)

        self.last_headers_name = None
        self.last_headers_value = None

        self.os_options = {
            'tenant_id': self.tenant_id,
            'tenant_name': self.tenant_name,
            'user_domain_id': self.user_domain_id,
            'user_domain_name': self.user_domain_name,
            'project_domain_id': self.project_domain_id,
            'project_domain_name': self.project_domain_name,
            'region_name': self.region_name,
        }
        self.os_options.update(self.os_extra_options)

        if not self.lazy_connect:
            self.swift_conn

    @property
    def swift_conn(self):
        """Get swift connection wrapper"""
        if not self._swift_conn:
            self._swift_conn = swiftclient.Connection(
                authurl=self.api_auth_url,
                user=self.api_username,
                key=self.api_key,
                retries=self.max_retries,
                tenant_name=self.tenant_name,
                os_options=self.os_options,
                auth_version=self.auth_version)
            self._check_container()
        return self._swift_conn

    def _check_container(self):
        """
        Check that container exists; raises exception if not.
        """
        try:
            self.swift_conn.head_container(self.container_name)
        except swiftclient.ClientException:
            headers = {}
            if self.auto_create_container:
                if self.auto_create_container_public:
                    headers['X-Container-Read'] = '.r:*'
                if self.auto_create_container_allow_orgin:
                    headers['X-Container-Meta-Access-Control-Allow-Origin'] = \
                        self.auto_create_container_allow_orgin
                self.swift_conn.put_container(self.container_name,
                                              headers=headers)
            else:
                raise ImproperlyConfigured(
                    "Container %s does not exist." % self.container_name)

    @property
    def base_url(self):
        if self._base_url is None:
            if self.auto_base_url:
                # Derive a base URL based on the authentication information from
                # the server, optionally overriding the protocol, host/port and
                # potentially adding a path fragment before the auth information.
                self._base_url = self.swift_conn.url + '/'
                if self.override_base_url is not None:
                    # override the protocol and host, append any path fragments
                    split_derived = urlparse.urlsplit(self._base_url)
                    split_override = urlparse.urlsplit(self.override_base_url)
                    split_result = [''] * 5
                    split_result[0:2] = split_override[0:2]
                    split_result[2] = (split_override[2] + split_derived[2]
                                       ).replace('//', '/')
                    self._base_url = urlparse.urlunsplit(split_result)

                self._base_url = urlparse.urljoin(self._base_url,
                                                  self.container_name)
                self._base_url += '/'
            else:
                self._base_url = self.override_base_url
        return self._base_url

    def _open(self, name, mode='rb'):
        original_name = name
        name = self.name_prefix + name

        headers, content = self.swift_conn.get_object(self.container_name, name)
        buf = BytesIO(content)
        buf.name = os.path.basename(original_name)
        buf.mode = mode
        return File(buf)

    def _save(self, name, content, headers=None):
        original_name = name
        name = self.name_prefix + name

        # Django rewinds file position to the beginning before saving,
        # so should we.
        # See django.core.files.storage.FileSystemStorage#_save
        # and django.core.files.base.File#chunks
        try:
            content.seek(0)
        except (AttributeError, UnsupportedOperation):  # pragma: no cover
            pass

        if self.content_type_from_fd:
            content_type = magic.from_buffer(content.read(1024), mime=True)
            # Go back to the beginning of the file
            content.seek(0)
        else:
            content_type = mimetypes.guess_type(name)[0]

        if self.content_length_from_fd:
            content_length = content.size
        else:
            content_length = None

        if content_type in self.gzip_content_types or (
           content_type is None and self.gzip_unknown_content_type):
            gz_data = BytesIO()
            gzf = gzip.GzipFile(filename=name,
                                fileobj=gz_data,
                                mode='wb',
                                compresslevel=self.gzip_compression_level)
            gzf.write(content.file.read())
            gzf.close()
            content = gz_data.getvalue()
            content_length = None

            if not headers:
                headers = {}
            headers['Content-Encoding'] = 'gzip'

        self.swift_conn.put_object(self.container_name,
                                   name,
                                   content,
                                   content_length=content_length,
                                   content_type=content_type,
                                   headers=headers)
        return original_name

    def get_headers(self, name):
        if self.cache_headers:
            """
            Optimization : only fetch headers once when several calls are made
            requiring information for the same name.
            When the caller is collectstatic, this makes a huge difference.
            According to my test, we get a *2 speed up. Which makes sense : two
            api calls were made..
            """
            if name != self.last_headers_name:
                # miss -> update
                self.last_headers_value = self.swift_conn.head_object(self.container_name, name)
                self.last_headers_name = name
        else:
            self.last_headers_value = self.swift_conn.head_object(self.container_name, name)
            self.last_headers_name = name

        return self.last_headers_value

    @prepend_name_prefix
    def exists(self, name):
        try:
            self.get_headers(name)
        except swiftclient.ClientException:
            return False
        return True

    @prepend_name_prefix
    def delete(self, name):
        try:
            self.swift_conn.delete_object(self.container_name, name)
        except swiftclient.ClientException:
            pass

    def get_valid_name(self, name):
        s = name.strip().replace(' ', '_')
        return re.sub(r'(?u)[^-_\w./]', '', s)

    @prepend_name_prefix
    def get_available_name(self, name, max_length=None):
        """
        Returns a filename that's free on the target storage system, and
        available for new content to be written to.
        """
        if not self.auto_overwrite:
            if max_length is None:
                name = super(SwiftStorage, self).get_available_name(name)
            else:
                name = super(SwiftStorage, self).get_available_name(
                    name, max_length)

        if self.name_prefix:
            # Split out the name prefix so we can just return the bit of
            # the name that's relevant upstream, since the prefix will
            # be automatically added on subsequent requests anyway.
            empty, prefix, final = name.partition(self.name_prefix)
            return final
        else:
            return name

    @prepend_name_prefix
    def size(self, name):
        return int(self.get_headers(name)['content-length'])

    @prepend_name_prefix
    def modified_time(self, name):
        return datetime.fromtimestamp(
            float(self.get_headers(name)['x-timestamp']))

    @prepend_name_prefix
    def url(self, name):
        return self._path(name)

    def _path(self, name):
        try:
            name = name.encode('utf-8')
        except UnicodeDecodeError:
            pass
        url = urlparse.urljoin(self.base_url, urlparse.quote(name))

        # Are we building a temporary url?
        if self.use_temp_urls:
            expires = int(time() + int(self.temp_url_duration))
            path = urlparse.unquote(urlparse.urlsplit(url).path)
            tmp_path = generate_temp_url(path, expires, self.temp_url_key, 'GET', absolute=True)
            url = urlparse.urljoin(self.base_url, tmp_path)

        return url

    def path(self, name):
        raise NotImplementedError

    @prepend_name_prefix
    def isdir(self, name):
        return '.' not in name

    @prepend_name_prefix
    def listdir(self, path):
        container = self.swift_conn.get_container(
            self.container_name, prefix=path, full_listing=self.full_listing)
        files = []
        dirs = []
        for obj in container[1]:
            remaining_path = obj['name'][len(path):].split('/')
            key = remaining_path[0] if remaining_path[0] else remaining_path[1]

            if not self.isdir(key):
                files.append(key)
            elif key not in dirs:
                dirs.append(key)

        return dirs, files

    @prepend_name_prefix
    def makedirs(self, dirs):
        self.swift_conn.put_object(self.container_name,
                                   '%s/.' % (self.name_prefix + dirs),
                                   contents='')

    @prepend_name_prefix
    def rmtree(self, abs_path):
        container = self.swift_conn.get_container(self.container_name)

        for obj in container[1]:
            if obj['name'].startswith(abs_path):
                self.swift_conn.delete_object(self.container_name,
                                              obj['name'])


class StaticSwiftStorage(SwiftStorage):
    container_name = setting('SWIFT_STATIC_CONTAINER_NAME', '')
    name_prefix = setting('SWIFT_STATIC_NAME_PREFIX', '')
    auto_base_url = setting('SWIFT_STATIC_AUTO_BASE_URL', True)
    override_base_url = setting('SWIFT_STATIC_BASE_URL')
    auto_create_container_public = True
    use_temp_urls = False

    def get_available_name(self, name, max_length=None):
        """
        When running collectstatic we don't want to return an available name,
        we want to return the same name because if the file exists we want to
        overwrite it.
        """
        return name




# -*- coding: UTF-8 -*-
import hmac
from copy import deepcopy
from django.test import TestCase
from django.core.exceptions import ImproperlyConfigured, SuspiciousFileOperation
from django.core.files.base import ContentFile
from hashlib import sha256
from mock import patch
from .utils import FakeSwift, auth_params, base_url, CONTAINER_CONTENTS, TENANT_ID
from swift import storage
from six.moves.urllib import parse as urlparse


class SwiftStorageTestCase(TestCase):

    def default_storage(self, auth_config, exclude=None, **params):
        """Instantiate default storage with auth parameters"""
        return storage.SwiftStorage(**auth_params(auth_config, exclude=exclude, **params))

    def static_storage(self, auth_config, exclude=None, **params):
        """Instantiate static storage with auth parameters"""
        return storage.StaticSwiftStorage(**auth_params(auth_config, exclude=exclude, **params))


@patch('swift.storage.swiftclient', new=FakeSwift)
class AuthTest(SwiftStorageTestCase):
    """Test authentication parameters"""

    def test_auth_v1(self):
        """Test version 1 authentication"""
        self.default_storage('v1')

    def test_auth_v2(self):
        """Test version 2 authentication"""
        self.default_storage('v2')

    def test_auth_v3(self):
        """Test version 3 authentication"""
        self.default_storage('v3')

    def test_auth_v1_detect_version(self):
        """Test version 1 authentication detection"""
        backend = self.default_storage('v1', exclude=['auth_version'])
        self.assertEqual(backend.auth_version, '1')

    def test_auth_v2_detect_version(self):
        """Test version 2 authentication detection"""
        backend = self.default_storage('v2', exclude=['auth_version'])
        self.assertEqual(backend.auth_version, '2')

    def test_auth_v3_detect_version(self):
        """Test version 3 authentication detection"""
        backend = self.default_storage('v3', exclude=['auth_version'])
        self.assertEqual(backend.auth_version, '3')

    def test_auth_v2_no_tenant(self):
        """Missing tenant in v2 auth"""
        with self.assertRaises(ImproperlyConfigured):
            self.default_storage('v2', exclude=['tenant_name', 'tenant_id'])

    def test_auth_v3_no_tenant(self):
        """Missing tenant in v3 auth"""
        with self.assertRaises(ImproperlyConfigured):
            self.default_storage('v3', exclude=['tenant_name'])

    def test_auth_v3_no_user_domain(self):
        """Missing user_domain in v3 auth"""
        with self.assertRaises(ImproperlyConfigured):
            self.default_storage('v3', exclude=['user_domain_name', 'user_domain_id'])

    def test_auth_v3_no_project_domain(self):
        """Missing project_domain in v3 auth"""
        with self.assertRaises(ImproperlyConfigured):
            self.default_storage('v3', exclude=['project_domain_name', 'project_domain_id'])

    def test_auth_v3_int_auth_version(self):
        """Auth version converts into a string"""
        backend = self.default_storage('v3', auth_version=3)
        self.assertEqual(backend.auth_version, '3')


@patch('swift.storage.swiftclient', new=FakeSwift)
class MandatoryParamsTest(SwiftStorageTestCase):

    def test_instantiate_default(self):
        """Instantiate default backend with no parameters"""
        with self.assertRaises(ImproperlyConfigured):
            storage.SwiftStorage()

    def test_instantiate_static(self):
        """Instantiate static backend with no parameters"""
        with self.assertRaises(ImproperlyConfigured):
            storage.StaticSwiftStorage()

    def test_mandatory_auth_url(self):
        """Test ImproperlyConfigured if api_auth_url is missing"""
        with self.assertRaises(ImproperlyConfigured):
            self.default_storage('v3', exclude=['api_auth_url'])

    def test_mandatory_username(self):
        """Test ImproperlyConfigured if api_auth_url is missing"""
        with self.assertRaises(ImproperlyConfigured):
            self.default_storage('v3', exclude=['api_username'])

    def test_mandatory_container_name(self):
        """Test ImproperlyConfigured if container_name is missing"""
        with self.assertRaises(ImproperlyConfigured):
            self.default_storage('v3', exclude=['container_name'])

    def test_mandatory_static_container_name(self):
        """Test ImproperlyConfigured if container_name is missing"""
        with self.assertRaises(ImproperlyConfigured):
            self.static_storage('v3', exclude=['container_name'])

    def test_mandatory_password(self):
        """Test ImproperlyConfigured if api_key is missing"""
        with self.assertRaises(ImproperlyConfigured):
            self.default_storage('v3', exclude=['api_key'])


@patch('swift.storage.swiftclient', new=FakeSwift)
class ConfigTest(SwiftStorageTestCase):

    def test_missing_container(self):
        """Raise if container don't exist"""
        with self.assertRaises(ImproperlyConfigured):
            self.default_storage('v3', container_name='idontexist')

    def test_delete_nonexisting_file(self):
        """Deleting non-existing file is silently ignored"""
        backend = self.default_storage('v3')
        backend.delete("idontexist.something")

    def test_auto_base_url(self):
        """Automatically resolve base url"""
        backend = self.default_storage('v3', auto_base_url=True)
        self.assertEqual(backend.base_url, base_url(container="container"))

    def test_override_base_url_no_auto(self):
        """Test overriding base url without auto base url"""
        url = 'http://localhost:8080/test/'
        backend = self.default_storage('v3',
                                       auto_base_url=False,
                                       override_base_url=url)
        self.assertEqual(backend.base_url, url)

    def test_override_base_url_auto(self):
        """Test overriding base url with auto"""
        url = 'http://localhost:8080'
        backend = self.default_storage('v3',
                                       auto_base_url=True,
                                       override_base_url=url)
        self.assertTrue(backend.base_url.startswith(url))
        storage_url = '{}/v1/AUTH_{}/{}/'.format(url, TENANT_ID, "container")
        self.assertEqual(backend.base_url, storage_url)

    def test_illegal_extra_opts(self):
        """extra_opts should always be a dict"""
        with self.assertRaises(ImproperlyConfigured):
            self.default_storage('v3', os_extra_options="boom!")

    @patch.object(FakeSwift.Connection, '__init__', return_value=None)
    def test_override_lazy_connect(self, mock_swift_init):
        """Test setting lazy_connect delays connection creation"""
        backend = self.default_storage('v3', lazy_connect=True)
        assert not mock_swift_init.called
        self.assertFalse(backend.exists('warez/some_random_movie.mp4'))
        assert mock_swift_init.called


# @patch('swift.storage.swiftclient', new=FakeSwift)
# class TokenTest(SwiftStorageTestCase):

#     def test_get_token(self):
#         """Renewing token"""
#         backend = self.default_storage('v3', auth_token_duration=0)
#         backend.get_token()

#     def test_set_token(self):
#         """Set token manually"""
#         backend = self.default_storage('v3', auth_token_duration=0)
#         backend.set_token('token')


@patch('swift.storage.swiftclient', new=FakeSwift)
class CreateContainerTest(SwiftStorageTestCase):

    def test_auto_create_container(self):
        """Auth create container"""
        self.default_storage(
            'v3',
            auto_create_container=True,
            auto_create_container_public=True,
            auto_create_container_allow_orgin=True,
            container_name='new')


@patch('swift.storage.swiftclient', new=FakeSwift)
class BackendTest(SwiftStorageTestCase):

    @patch('swift.storage.swiftclient', new=FakeSwift)
    def setUp(self):
        self.backend = self.default_storage('v3')

    def test_url(self):
        """Get url for a resource"""
        name = 'images/test.png'
        url = self.backend.url(name)
        self.assertEqual(url, base_url(container=self.backend.container_name, path=name))

    def test_url_unicode_name(self):
        """Get url for a resource with unicode filename"""
        name = u'images/test终端.png'
        url = self.backend.url(name)
        self.assertEqual(url, base_url(container=self.backend.container_name, path=name))

    def test_object_size(self):
        """Test getting object size"""
        size = self.backend.size('images/test.png')
        self.assertEqual(size, 4096)

    def test_modified_time(self):
        """Test getting modified time of an object"""
        self.backend.modified_time('images/test.png')

    def test_object_exists(self):
        """Test for the existence of an object"""
        exists = self.backend.exists('images/test.png')
        self.assertTrue(exists)

    def test_object_dont_exists(self):
        """Test for the existence of an non-existent object"""
        exists = self.backend.exists('warez/some_random_movie.mp4')
        self.assertFalse(exists)

    def test_get_headers_chache1(self):
        self.backend.cache_headers = True;
        headers = self.backend.get_headers('images/test.png')
        self.assertEqual('fcfc6539ce4e545ce58bafeeac3303a7', headers['hash'])

    def test_get_headers_chache2(self):
        self.backend.cache_headers = False;
        headers = self.backend.get_headers('images/test.png')
        self.assertEqual('fcfc6539ce4e545ce58bafeeac3303a7', headers['hash'])

    def test_listdir(self):
        """List root in container"""
        dirs, files = self.backend.listdir('')
        self.assertListEqual(dirs, ['images', 'css', 'js'])
        self.assertListEqual(files, ['root.txt'])

    @patch('tests.utils.FakeSwift.objects', new=deepcopy(CONTAINER_CONTENTS))
    def test_rmtree(self):
        """Remove folder in storage"""
        backend = self.default_storage('v3')
        backend.rmtree('images')
        dirs, files = self.backend.listdir('')
        self.assertListEqual(dirs, ['css', 'js'])
        self.assertListEqual(files, ['root.txt'])

    @patch('tests.utils.FakeSwift.objects', new=deepcopy(CONTAINER_CONTENTS))
    def test_mkdirs(self):
        """Make directory/pseudofolder in backend"""
        backend = self.default_storage('v3')
        backend.makedirs('downloads')
        dirs, files = self.backend.listdir('')
        self.assertListEqual(dirs, ['images', 'css', 'js', 'downloads'])
        self.assertListEqual(files, ['root.txt'])

    @patch('tests.utils.FakeSwift.objects', new=deepcopy(CONTAINER_CONTENTS))
    def test_delete_object(self):
        """Delete an object"""
        backend = self.default_storage('v3')
        backend.delete('root.txt')
        dirs, files = self.backend.listdir('')
        self.assertListEqual(dirs, ['images', 'css', 'js'])
        self.assertListEqual(files, [])

    @patch('tests.utils.FakeSwift.objects', new=deepcopy(CONTAINER_CONTENTS))
    def test_save(self):
        """Save an object"""
        backend = self.default_storage('v3')
        content_file = ContentFile("Hello world!")
        name = backend.save("test.txt", content_file)
        dirs, files = self.backend.listdir('')
        self.assertEqual(files.count(name), 1)

    @patch('tests.utils.FakeSwift.objects', new=deepcopy(CONTAINER_CONTENTS))
    @patch('gzip.GzipFile')
    def test_save_gzip(self, gzip_mock):
        """Save an object"""
        backend = self.default_storage('v3')
        backend.gzip_content_types = ['text/plain']
        content_file = ContentFile(b'Hello world!')
        name = backend.save('testgz.txt', content_file)
        dirs, files = self.backend.listdir('')
        self.assertEqual(files.count(name), 1)
        self.assertTrue(gzip_mock.called)

    @patch('tests.utils.FakeSwift.objects', new=deepcopy(CONTAINER_CONTENTS))
    def test_content_type_from_fd(self):
        """Test content_type detection on save"""
        backend = self.default_storage('v3', content_type_from_fd=True)
        backend.save("test.txt", ContentFile("Some random data"))

    def test_save_non_rewound(self):
        """Save file with position not at the beginning"""
        content = dict(orig=b"Hello world!")
        content_file = ContentFile(content['orig'])
        content_file.seek(5)

        def mocked_put_object(cls, url, token, container, name=None,
                              contents=None, content_length=None, *args, **kwargs):
            content['saved'] = contents.read()
            content['size'] = content_length

        with patch('tests.utils.FakeSwift.put_object', new=classmethod(mocked_put_object)):
            self.backend.save('test.txt', content_file)
        self.assertEqual(content['saved'], content['orig'])
        self.assertEqual(content['size'], len(content['orig']))

    def test_no_content_length_from_fd(self):
        """Test disabling content_length_from_fd on save"""
        backend = self.default_storage('v3', content_length_from_fd=False)
        content = dict(orig="Hello world!")
        content_file = ContentFile("")
        content_file.write(content['orig'])

        def mocked_put_object(cls, url, token, container, name=None,
                              contents=None, content_length=None, *args, **kwargs):
            content['saved'] = contents.read()
            content['size'] = content_length

        with patch('tests.utils.FakeSwift.put_object', new=classmethod(mocked_put_object)):
            backend.save('test.txt', content_file)
        self.assertEqual(content['saved'], content['orig'])
        self.assertIsNone(content['size'])

    def test_open(self):
        """Attempt to open a object"""
        file = self.backend._open('root.txt')
        self.assertEqual(file.name, 'root.txt')
        data = file.read()
        self.assertEqual(len(data), 4096)

    def test_get_available_name_nonexist(self):
        """Available name for non-existent object"""
        object = 'images/doesnotexist.png'
        name = self.backend.get_available_name(object)
        self.assertEqual(name, object)

    def test_get_available_name_max_length_nonexist(self):
        """Available name with max_length for non-existent object"""
        object = 'images/doesnotexist.png'
        name = self.backend.get_available_name(object, len(object))
        self.assertEqual(name, object)
        with self.assertRaises(SuspiciousFileOperation):
            name = self.backend.get_available_name(object, 16)

    def test_get_available_name_exist(self):
        """Available name for existing object"""
        object = 'images/test.png'
        name = self.backend.get_available_name(object)
        self.assertNotEqual(name, object)

    def test_get_available_name_max_length_exist(self):
        """Available name with max_length for existing object"""
        object = 'images/test.png'
        name = self.backend.get_available_name(object, 32)
        self.assertNotEqual(name, object)

    def test_get_available_name_prefix(self):
        """Available name with prefix"""
        object = 'test.png'
        # This will add the prefix, then get_avail will remove it again
        backend = self.default_storage('v3', name_prefix="prefix-")
        name = backend.get_available_name(object)
        self.assertEqual(object, name)

    def test_get_available_name_static(self):
        """Static's get_available_name should be unmodified"""
        static = self.static_storage('v3')
        object = "test.txt"
        name = static.get_available_name(object)
        self.assertEqual(name, object)

    def test_get_valid_name(self):
        name = self.backend.get_valid_name("A @#!file.txt")
        self.assertEqual(name, "A_file.txt")

    def test_path(self):
        """path is not implemented"""
        with self.assertRaises(NotImplementedError):
            self.backend.path("test.txt")


@patch('swift.storage.swiftclient', new=FakeSwift)
class TemporaryUrlTest(SwiftStorageTestCase):

    def assert_valid_temp_url(self, name):
        url = self.backend.url(name)
        split_url = urlparse.urlsplit(url)
        query_params = urlparse.parse_qs(split_url[3])
        split_base_url = urlparse.urlsplit(base_url(container=self.backend.container_name, path=name))

        # ensure scheme, netloc, and path are same as to non-temporary URL
        self.assertEqual(split_base_url[0:2], split_url[0:2])

        # ensure query string contains signature and expiry
        self.assertIn('temp_url_sig', query_params)
        self.assertIn('temp_url_expires', query_params)

    def assert_valid_signature(self, path):
        """Validate temp-url signature"""
        backend = self.default_storage('v3', use_temp_urls=True, temp_url_key='Key')
        url = backend.url(path)
        url_parsed = urlparse.urlsplit(url)
        params = urlparse.parse_qs(url_parsed.query)
        msg = "{}\n{}\n{}".format("GET", params['temp_url_expires'][0], urlparse.unquote(url_parsed.path))
        sig = hmac.new(backend.temp_url_key, msg.encode('utf-8'), sha256).hexdigest()
        self.assertEqual(params['temp_url_sig'][0], sig)

    def test_signature(self):
        self.assert_valid_signature("test/test.txt")
        self.assert_valid_signature("test/file with spaces.txt")

    def test_temp_url_key_required(self):
        """Must set temp_url_key when use_temp_urls=True"""
        with self.assertRaises(ImproperlyConfigured):
            self.backend = self.default_storage('v3', use_temp_urls=True)

    def test_temp_url_key(self):
        """Get temporary url using string key"""
        self.backend = self.default_storage('v3', use_temp_urls=True, temp_url_key='Key')
        self.assert_valid_temp_url('images/test.png')

    def test_temp_url_key_unicode(self):
        """temp_url_key must be ascii"""
        with self.assertRaises(ImproperlyConfigured):
            self.backend = self.default_storage('v3', use_temp_urls=True, temp_url_key=u'aあä')

    def test_temp_url_key_unicode_latin(self):
        """Get temporary url using a unicode key which can be ascii-encoded"""
        self.backend = self.default_storage('v3', use_temp_urls=True, temp_url_key=u'Key')
        self.assert_valid_temp_url('images/test.png')

    def test_temp_url_unicode_name(self):
        """temp_url file name can be unicode."""
        self.backend = self.default_storage('v3', use_temp_urls=True, temp_url_key=u'Key')
        self.assert_valid_temp_url('images/aあä.png')



from setuptools import setup

setup(
    name='django-storage-swift',
    version='1.4.0',
    description='OpenStack Swift storage backend for Django',
    long_description=open('README.rst').read(),
    url='https://github.com/dennisv/django-storage-swift',
    author='Dennis Vermeulen',
    author_email='blacktorn@gmail.com',
    license='MIT',
    packages=['swift'],
    install_requires=[
        'python-swiftclient>=2.2.0',
        'python-keystoneclient>=0.2.3',
        'six',
        'python-magic>=0.4.10',
    ],
    zip_safe=False,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
)
